# distutils: language=c++

from libcpp cimport bool
from libcpp.string cimport string

cdef extern from "eudist_cpp.hxx":
    double dot_dot(double*, double*, int)
    cppclass Plane:
        Plane(double*, double*, double*, int)
        double dist(double *)
        double signed_dist(double *)
        void info()
    int winding_number(double*, double*, int)
    double polygon_dot(double*, double* , int, int, bool)

    cppclass PolyMesh:
        PolyMesh(double*, double*, int, int)
        int find_cell(double *, int)



from eudist import *


def test_plane_point1():
    p0 = np.array([0.0, 0, 0], dtype=float)
    p1 = np.array([1.0, 0, 0], dtype=float)
    p2 = np.array([0.0, 1, 0], dtype=float)
    pl = Plane(p0=p0, p1=p1, p2=p2)
    for pnt, val in [
        (np.array([0.0, 0, 1]), 1),
        (np.array([0.0, 1, 1]), 1),
        (np.array([1.0, 0, 1]), 1),
        (np.array([0.0, 0, 2]), 2),
        (np.array([0.0, 0, 0]), 0),
    ]:
        assert plane_dot(pl, pnt) == val
        assert pl.dist(pnt) == val


def test_plane_point2():
    p0 = np.array([0.0, 0, 1], dtype=float)
    p1 = np.array([1.0, 0, 0], dtype=float)
    p2 = np.array([0.0, 1, 0], dtype=float)
    pl = Plane(p0=p0, p1=p1, p2=p2)
    assert plane_dot(pl, np.array([0.0, 0, 1])) == 0
    assert np.isclose(plane_dot(pl, np.array([0.0, 1, 1])), np.sqrt(1 / 3))
    assert np.isclose(plane_dot(pl, np.array([1.0, 0, 1])), np.sqrt(1 / 3))
    assert np.isclose(plane_dot(pl, np.array([0.0, 0, 0])), np.sqrt(1 / 3))


def random_rotation(dims):
    if dims == 3:
        trans = np.eye(3)
        for i, j in [[0, 1], [0, 2], [1, 2]]:
            rot = np.eye(3)
            phi = np.random.random() * np.pi * 2
            c = np.cos(phi)
            s = np.sin(phi)
            rot[i, i] = c
            rot[j, j] = c
            rot[i, j] = -s
            rot[j, i] = s
            trans = np.matmul(trans, rot)
        eig = np.abs(np.linalg.eig(trans)[0])
        # matrix is singular
        if np.max(eig) / np.min(eig) > 1 + dims * 1e-6:
            raise RuntimeError("Roational matrix should have eigenvalue 1!")
        return trans
    elif dims == 2:
        phi = np.random.random() * np.pi * 2
        c = np.cos(phi)
        s = np.sin(phi)
        rot = np.array([[c, -s], [s, c]])
        return rot
    elif dims > 3:
        trans = np.eye(dims)
        for i in range(dims):
            for j in range(i):
                rot = np.eye(dims)
                phi = np.random.random() * np.pi * 2
                c = np.cos(phi)
                s = np.sin(phi)
                rot[i, i] = c
                rot[j, j] = c
                rot[i, j] = -s
                rot[j, i] = s
                trans = np.matmul(trans, rot)
        eig = np.abs(np.linalg.eig(trans)[0])
        if np.max(eig) / np.min(eig) > 1 + dims * 1e-6:
            raise RuntimeError("Roational matrix should have eigenvalue 1!")
        return trans
    raise RuntimeError(f"not implemented for dim={dims}")


def test_plane_point3():
    for i in range(100):
        dist = 3
        ps = [
            np.array([0.0, 0, 0]),
            np.array([1.0, 0, 0]),
            np.array([0.0, 1, 0]),
            np.array([0.0, 0, dist]),
        ]

        ps[3][0:2] = np.random.normal(size=2)
        off1 = np.random.normal(scale=3, size=3)
        off2 = np.random.normal(scale=3, size=3)
        trans = random_rotation(dims=3)
        scal = np.random.lognormal()
        dist *= scal
        for p in ps:
            p *= scal
            p += off1
            p = np.matmul(trans, p)
            p += off2
        pl = Plane(p0=ps[0], p1=ps[1], p2=ps[2])
        assert np.isclose(plane_dot(pl, ps[3]), dist)
        assert np.isclose(pl.dist(ps[3]), dist)


def test_plane_point4():
    for i in range(100):
        ps = np.random.random((4, 3))
        ps[:3, 0] = 0
        dist = ps[3, 0]
        ps = [p for p in ps]
        off1 = np.random.normal(scale=3, size=3)
        off2 = np.random.normal(scale=3, size=3)
        trans = random_rotation(dims=3)
        scal = np.random.lognormal()
        dist *= scal
        for p in ps:
            p *= scal
            p += off1
            p = np.matmul(trans, p)
            p += off2
        pl = Plane(p0=ps[0], p1=ps[1], p2=ps[2])
        assert np.isclose(plane_dot(pl, ps[3]), dist)
        assert np.isclose(pl.dist(ps[3]), dist)


def test_plane_point5():
    ps = np.array(
        [
            [5.45458, 0.0, -0.374516],
            [5.47951, 0.0, -0.383842],
            [5.48446459, 0.01914447, -0.383323],
            [5.45953074, 0.01905743, -0.374144],
        ]
    )
    # ps = [p for p in ps]
    pl = Plane(p0=ps[0], p1=ps[1], p2=ps[2])
    assert pl.dist(ps[3]) < 1
    assert Plane(p0=ps[0], p1=ps[1], p2=ps[2]).dist(ps[3]) < 0.01


def test_dot_dot():
    for i in range(100):
        for n in [1, 2, 3, 5, 8]:
            p0 = np.random.normal(scale=3, size=n)
            p1 = np.random.normal(scale=3, size=n)
            assert np.isclose(dot_dot(p0, p1), np.sqrt(np.dot(p0 - p1, p0 - p1)))
    for i in range(100):
        n = 3
        rotate = random_rotation(n)
        p0 = np.zeros(n)
        p1 = np.zeros(n)
        p0[0] = 1
        p1[0] = 2
        p0 = p0 @ rotate
        p1 = p1 @ rotate
        assert np.isclose(dot_dot(p0, p1), 1)
    p0 = np.array([1.0, 0, 0])
    p1 = np.array([2.0, 0, 0])
    assert np.isclose(dot_dot(p0, p1), 1)
    p1 = np.array([0.0, 0, 0])
    assert np.isclose(dot_dot(p0, p1), 1)
    p0 = np.array([1.0])
    p1 = np.array([2.0])
    assert np.isclose(dot_dot(p0, p1), 1)
    p0 = np.array([1.0, 1, 1])
    p1 = np.array([2.0, 2, 2])
    assert np.isclose(dot_dot(p0, p1), np.sqrt(3))


def test_winding_number():
    for i in range(100):
        n = 2
        pnts = np.zeros((4, n))
        pnts[1, 0] = 1
        pnts[2, 0] = 1
        pnts[2, 1] = 1
        pnts[3, 1] = 1
        dot = np.random.normal(scale=3, size=n)
        expected = 0
        if 0 <= dot[0] <= 1 and 0 <= dot[1] <= 1:
            expected = 1
        rot = random_rotation(2)
        scal = np.random.lognormal()
        trans = rot * scal
        pnts = pnts @ trans
        dot = dot @ trans

        wn = winding_number(pnts, dot)
        assert wn == expected


def test_polygone_dot():
    for i in range(1000):
        n = 3
        # make a square
        pnts = np.zeros((4, n))
        pnts[1, 0] = 1
        pnts[2, 0] = 1
        pnts[2, 1] = 1
        pnts[3, 1] = 1
        pnts[:, :2] -= 0.5
        # A random dot
        dot = np.random.normal(scale=3, size=n)
        # dot = np.array([.4,.2,1])
        # Calculate distance
        dist = np.abs(dot)
        for i in range(2):
            dist[i] -= 0.5
            if dist[i] < 0:
                dist[i] = 0
        dist = np.sqrt(np.dot(dist, dist))
        # Add some offset
        offset = np.random.normal(scale=3, size=n)
        pnts += offset
        dot += offset

        # Rotate and scale
        rot = random_rotation(n)
        scal = np.random.lognormal()
        dist *= scal
        trans = rot * scal
        pnts = pnts @ trans
        dot = dot @ trans

        calc = polygon_dot(pnts, dot)
        if not np.isclose(dist, calc):
            import matplotlib.pyplot as plt

            for i in range(n):
                for j in range(i):
                    plt.figure()
                    plt.plot(pnts[:, i], pnts[:, j], "-o")
                    plt.plot(dot[i], dot[j], "x")
            plt.show()
            raise AssertionError


if __name__ == "__main__":
    test_plane_point1()
    test_plane_point2()
    test_plane_point3()
    test_plane_point4()
    test_dot_dot()
    test_winding_number()
    test_polygone_dot()



# distutils: language=c++
# distutils: include_dirs =
# distutils: libraries =
# distutils: library_dirs =
# distutils: sources = eudist_cpp.cxx
# distutils: extra_compile_args = -std=c++11

cimport eudist_cpp as c
cimport numpy as np
import numpy as np

def dot_dot(np.ndarray[double,ndim=1] a, np.ndarray[double, ndim=1] b):
    a = np.ascontiguousarray(a)
    b = np.ascontiguousarray(b)
    return c.dot_dot(&a[0],&b[0],len(a))


cdef class Plane:
    """
    Defines a plane.

    The constructor takes 3 points that are on the plane and span the plane.
    """
    cdef c.Plane * cobj
    def __cinit__(self, np.ndarray[double,ndim=1] p0, np.ndarray[double,ndim=1] p1, np.ndarray[double,ndim=1] p2):
        p0 = np.ascontiguousarray(p0)
        p1 = np.ascontiguousarray(p1)
        p2 = np.ascontiguousarray(p2)
        self.cobj = new c.Plane(&p0[0], &p1[0], &p2[0], len(p0))

    def __dealloc__(self):
        self.__mydealloc__()

    def __mydealloc__(self):
        if self.cobj:
            del self.cobj
            self.cobj = NULL

    def dist(self, np.ndarray[double,ndim=1] dot):
        dot = np.ascontiguousarray(dot)
        return self.cobj.dist(&dot[0])

    def signed_dist(self, np.ndarray[double,ndim=1] dot):
        dot = np.ascontiguousarray(dot)
        return self.cobj.signed_dist(&dot[0])

    def info(self):
        self.cobj.info()

cdef class PolyMesh:
    """
    Defines a mesh in 2 dimension, which allows for fast finding of in which cell a point is.

    The constructor takes two 2D array for the x and y position of the grid corners.
    """
    cdef c.PolyMesh * cobj
    def __cinit__(self, np.ndarray[double,ndim=2] datax, np.ndarray[double,ndim=2] datay):
        for i in range(2):
            if not datax.shape[i] == datay.shape[i]:
                xs = datax.shape
                ys = datay.shape
                raise ValueError(f"Data mismatch: ({xs[0]}, {xs[1]}) != ({ys[0]}, {ys[1]})")
        cdef np.ndarray[double, ndim=1, mode='c'] datax_ = np.ravel(datax, order='c')
        cdef np.ndarray[double, ndim=1, mode='c'] datay_ = np.ravel(datay, order='c')
        self.cobj = new c.PolyMesh(&datax_[0], &datay_[0], datax.shape[0], datay.shape[1])

    def __dealloc__(self):
        self.__mydealloc__()

    def __mydealloc__(self):
        if self.cobj:
            del self.cobj
            self.cobj = NULL

    def find_cell(self,  np.ndarray[double,ndim=1] dot, int guess=-1):
        """
        Find the cellid of a point.
        """
        dot = np.ascontiguousarray(dot)
        return self.cobj.find_cell(&dot[0], guess)

def plane_dot(Plane pl, np.ndarray[double, ndim=1] dot):
    return pl.dist(dot)

def winding_number(np.ndarray[double,ndim=2] points, np.ndarray[double, ndim=1] dot):
    """
    Calculate the winding number between the 2-d surface spawened by
    the `points` and the `dot`.  A winding number of zero means the
    point is outside the surface.

    If the conversions are slow, ensure the variables are already
    C-contigous.
    """
    cdef np.ndarray[double, ndim=1, mode='c'] pnts = np.ravel(points,order='c')
    dot = np.ascontiguousarray(dot)
    return c.winding_number(&pnts[0], &dot[0], len(points))


def polygon_dot(np.ndarray[double,ndim=2] points, np.ndarray[double, ndim=1] dot, check_planar=False):
    """
    Calculate the distance between a polygon and a dot.

    The polygon is assumed to be flat.
    """
    cdef np.ndarray[double, ndim=1, mode='c'] pnts = np.ravel(points,order='c')
    dot = np.ascontiguousarray(dot)
    return c.polygon_dot(&pnts[0], &dot[0], len(points), len(dot), check_planar)



def det(a, b):
    return a[0] * b[1] - a[1] * b[0]


def vlen(a):
    return np.sqrt(np.dot(a, a))


def do_seg_seg_intersect(seq0, seq1):
    """
    Check if the two line segments intersect.
    """
    eps = 1e-16
    v0 = seq0[1] - seq0[0]
    v1 = seq1[1] - seq1[0]
    p0 = seq0[0]
    p1 = seq1[0]
    dist = p1 - p0

    det1 = det(v0, v1)
    # check for co-linearity:
    if abs(det1) < eps:
        # Check for overlap? - not yet
        # dist -= np.dot(dist, v0) / vlen(v0) / vlen(dist)
        # print("colin", dist, det1)
        return False

    t0 = det(dist, v0) / det1

    if t0 < -eps or t0 > 1 + eps:
        # print("t0", t0)
        return False

    t1 = det(dist, v1) / det1

    if t1 < -eps or t1 > 1 + eps:
        # print("t1", t1)
        return False

    return True



import eudist as eudist
import numpy as np


def random_rotation(dims):
    trans = np.eye(dims)
    for i in range(dims):
        for j in range(i):
            rot = np.eye(dims)
            phi = np.random.random() * np.pi * 2
            c = np.cos(phi)
            s = np.sin(phi)
            rot[i, i] = c
            rot[j, j] = c
            rot[i, j] = -s
            rot[j, i] = s
            trans = np.matmul(trans, rot)
    eig = np.abs(np.linalg.eig(trans)[0])
    if np.max(eig) / np.min(eig) > 1 + dims * 1e-6:
        raise RuntimeError("Roational matrix should have eigenvalue 1!")
    return trans


def dotestseqseq(a, b, *, iter=100, expect=True):
    assert eudist.do_seg_seg_intersect(a, b) == expect
    a0 = a
    b0 = b
    for i in range(iter):
        a = a0.copy()
        b = b0.copy()
        off1 = np.random.normal(scale=3, size=2)
        off2 = np.random.normal(scale=3, size=2)
        trans = random_rotation(dims=2)
        scal = np.random.lognormal()
        for p in [a, b]:
            p *= scal
            p += off1
            p = np.matmul(trans, p)
            p += off2
        assert eudist.do_seg_seg_intersect(a, b) == expect


def test_seq_seq0():
    a = np.zeros((2, 2))
    b = np.zeros((2, 2))
    a[0, 1] = 1
    a[1, 1] = -1
    b[0, 0] = 1
    b[1, 0] = -1
    dotestseqseq(a, b)
    dotestseqseq(b, a)


test_seq_seq0()


def test_seq_seq1():
    a = np.zeros((2, 2))
    b = np.zeros((2, 2))
    a[0, 1] = 1
    b[1, 0] = 1
    dotestseqseq(a, b)
    dotestseqseq(b, a)


test_seq_seq1()


def test_seq_seq2():
    A = [[0.0, 0], [3, 3]]
    B = [[-2.0, 2], [2, -2]]
    for _ in range(10):
        a = np.array(A)
        b = np.array(B)
        off = np.random.random(1) * 5

        dotestseqseq(a, b + off, iter=10, expect=(off < 3))


def test_seq_seq3():
    for i in [1, -1]:
        for j in [1, -1]:
            a = np.array([[0.0, 0], [0, 1]])[::i]
            b = np.array([[0.0, 0], [1, 0]])[::j]
            dotestseqseq(a, b, iter=10)


# def test_parallel():
#     a = np.array([[0.0, 0], [0, 1]])
#     for i in range(10):
#         b = a.copy()
#         off = np.random.random(1) * 2
#         b[:, 1] += off
#         dotestseqseq(a, b, iter=10, expect=(off < 1))

# def test_seq_seq4():
#     a = np.array([[0.0, 0], [0, 1]])
#     for i in range(10):
#         b = np.array([[0.0, 0], [0, 1]])
#         off = np.random.random(1) * 2
#         b[:, 1] += off
#         dotestseqseq(a, b, iter=10, expect=(off < 1))



from setuptools import Extension, setup

incdir = ""
try:
    from Cython.Build import cythonize
    import numpy as np

    incdir = np.get_include()
except ImportError:
    # Ignore error, as otherwise setuptools doesn't know we need cython
    def cythonize(*args, **kwargs):
        pass


sourcefiles = ["eudist.pyx", "eudist_cpp.cxx"]

extensions = [Extension("eudist", sourcefiles, include_dirs=[incdir])]

setup(
    use_scm_version=True,
    ext_modules=cythonize(
        extensions,
        language_level=3,
        compiler_directives=dict(binding=True, embedsignature=True),
    ),
)



import numpy as np
import eudist
from test_point_plane import random_rotation


class Transform(object):
    def __init__(self, dims=2):
        self.rot = random_rotation(dims)
        self.dims = dims
        self.off1 = np.random.normal(scale=3, size=dims)
        self.off2 = np.random.normal(scale=3, size=dims)
        self.scale = np.random.lognormal()

    def __call__(self, p):
        # return p
        slc = slice(None), None
        shp = p.shape
        p = p.reshape(2, -1)
        p *= self.scale
        p += self.off1[slc]
        p = np.matmul(self.rot, p)
        p += self.off2[slc]
        p = p.reshape(shp)
        return p


def test_polymesh_1():
    nx = 5
    ny = 7
    x = np.arange(nx + 1)[:, None] * np.ones((nx + 1, ny + 1))
    y = np.arange(ny + 1)[None, :] * np.ones((nx + 1, ny + 1))
    mesh = np.array([x, y])  # .flatten(), y.flatten()])
    trans = Transform(2)
    mesh = trans(mesh)
    cx, cy = np.random.randint(nx), np.random.randint(ny)
    cell0 = np.array([[cx, cy], [cx, cy + 1.0], [cx + 1, cy + 1], [cx + 1, cy]]).T
    cellid = cx * ny + cy
    cell = trans(cell0)
    a = cell[:, 0]
    b = cell[:, 1] - a
    c = cell[:, 3] - a
    d = cell[:, 2] - a - b - c
    meshx, meshy = mesh
    mesh = eudist.PolyMesh(meshx, meshy)

    for _ in range(10):
        xx, xy = np.random.rand(2) * 1.5 - 0.25
        pos = a + b * xx + c * xy + xx * xy * d
        isin = 0 <= xx <= 1 and 0 <= xy <= 1
        try:
            assert (mesh.find_cell(pos) == cellid) == isin
        except AssertionError:
            print(mesh.find_cell(pos), cellid, cx, cy, nx, ny)
            import matplotlib.pyplot as plt

            print(meshx.shape)
            data = ((x == cx) & (y == cy)).astype(int)
            plt.pcolormesh(meshx, meshy, data)
            plt.scatter(pos[0], pos[1])
            plt.plot(cell[0], cell[1])
            plt.show()
            raise



"""
Library for calculating euclidean distances.

    This file is part of eudist.

    eudist is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    eudist is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with eudist.  If not, see <https://www.gnu.org/licenses/>.

"""

import numpy as np


class Plane(object):
    def __init__(self, *, p0, p1=None, p2=None):
        """
        The plane is either defined by 3 points, a point and a normal
        vector, or a point and two span vectors.
        Only 3 points are currently supported.
        """
        self.p0 = p0
        self.p1 = p1
        self.p2 = p2
        self.v0 = p1 - p0
        self.v1 = p2 - p0
        # [a, b, c] are the elements of norm
        self.norm = np.cross(self.v0, self.v1)
        self.d = -np.dot(self.norm, self.p0)

    def project(self, dot):
        """
        Project a dot onto this plane
        """
        if len(dot) == 2:
            return dot
        return dot - self.norm * (np.dot(dot, self.norm) + self.d) / (
            np.dot(self.norm, self.norm)
        )

    def dist(self, dot):
        return dist_plane_dot(self, dot)


def dist_plane_dot(plane, dot):
    if len(dot) == 2:
        return 0
    return np.abs(np.dot(dot, plane.norm) + plane.d) / np.sqrt(
        np.dot(plane.norm, plane.norm)
    )


def dist_dot_dot(p0, p1):
    """
    Distance between two points
    """
    dist = p0 - p1
    return np.sqrt(np.dot(dist, dist))


def dist_line_segment_dot(line, dot):
    """
    Calculate the distance between the line segment defined by two
    points in line and a dot. 
    """
    v = line[1] - line[0]
    w = dot - line[0]

    c1 = np.dot(w, v)
    if c1 <= 0:
        return dist_dot_dot(dot, line[0])

    c2 = np.dot(v, v)
    if c2 <= c1:
        return dist_dot_dot(dot, line[1])

    p = c1 / c2
    proj = line[0] + p * v
    return dist_dot_dot(proj, dot)


def winding_number(points, dot):
    """
    Calculate winding number.
    
    Based on https://geomalgorithms.com/a03-_inclusion.html
    by softSurfer and Dan Sunday
    """

    def is_left(p0, p1, p2):
        d1 = p1 - p0
        d2 = p2 - p0
        return (d1[0] * d2[1]) - (d2[0] * d1[1])

    # Winding number counter
    wn = 0
    for i in range(len(points)):
        # Use -1 to get periodic boundary conditions
        if points[i - 1][1] <= dot[1]:
            if points[i][1] > dot[1]:
                if is_left(points[i - 1], points[i], dot) > 0:
                    wn += 1
        else:
            if points[i][1] <= dot[1]:
                if is_left(points[i - 1], points[i], dot) < 0:
                    wn -= 1
    return wn


def is_planar(points, *, rtol=1e-3, atol=1e-8):
    # 3 points are alwys in a plane
    if len(points) < 4:
        return True
    # 2 Dimensions are always in a plane
    if len(points[0]) < 3:
        return True
    scale = dist_dot_dot(points[0], points[1]) + dist_dot_dot(points[0], points[2])
    scale /= 2
    plane = Plane(p0=points[0], p1=points[1], p2=points[2])
    for i in range(3, len(points)):
        if plane.dist(points[i]) > scale * 1e-3:
            return False
    return True


def dist_polygon_dot(points, dot, check_planar=True):
    if len(points) == 1:
        return dist_dot_dot(points[0], dot)
    elif len(points) == 2:
        return dist_line_segment_dot(points, dot)
    plane = Plane(p0=points[0], p1=points[1], p2=points[2])
    if check_planar:
        if len(points) > 3 and len(dot) > 2:
            # get an estimate of the length scales involved
            if not is_planar(points, atol=0, rtol=1e-3):
                raise RuntimeError(f"Point of polygon are not in a plane!")

    # Simple projection onto 2D-plane. Drop main component of orthogonal vector
    if len(dot) == 3:
        slcr = [0, 1, 2]
        slcr.pop(np.argmax(np.abs(plane.norm)))
    elif len(dot) == 2:
        slcr = slice(None)
    else:
        raise RuntimeError("Only 2D or 3D supported!")

    wn = winding_number([p[slcr] for p in points], plane.project(dot)[slcr])
    if wn == 0:
        return np.min(
            [
                dist_line_segment_dot([points[i - 1], points[i]], dot)
                for i in range(len(points))
            ]
        )
    return plane.dist(dot)
